#include "columns.h"

#include "justifyCommon.h"
#include "stringHelper.h"

#include <array>
#include <memory>
#include <string>
#include <vector>

/*
 Possible layout and what the variables mean:

  columnWidth columnHorizontalSpacing
  /         \  /\
 /           \/  \
 BLA BLA BLAAA    ASDFASDFASDFA  \
 BLA BLA BLA..    ASDF ASDF ASD   \ row = a vector of columns that all start
 BLABLABLABLA.    ASDASDASDASDA   /       on the same line, but may end on 
 BLABLA           FILLERTEXTLAL  /        a different line.
 \           /    ASDASD        /
  \         /                   \
     block                       \ columnVerticalSpacing
                                 / 
 MORE TEXT !!!    FOO FOO BAR.. \
 ASDF ASDF ASD    Fin.           \ another row
 ASD FOO.                        /

 An individual row can not contain more LINES than maxColumnHeight.

*/

// A block is a set of lines.
using block = std::vector<std::wstring>;
// ... but we also deal with sets of blocks. (eg, for representing a row)
using blocks = std::vector<block>;

std::vector<std::wstring> columns(
    const std::vector<std::wstring>& in,
    size_t numCols,
    size_t columnWidth,
    size_t columnHorizontalSpacing,
    size_t columnVerticalSpacing,
    size_t maxColumnHeight,
    size_t width,
    wchar_t pageSepLineChar)
{
    std::vector<std::wstring> result;

    // Basic sanity check.
    if ((columnWidth == 0) || (maxColumnHeight == 0)) return result;

    // Puncutation/characters that we can potentially consider a good spot
    // to start a new block on if a line ends in it.
    static std::array<wchar_t, 10> const blockEnders = {
        '!', '?', '.', ';', '}', ']', '>'
    };

    // The input text (in variable) should already be justified and have a
    // max text width per line of columnWidth.

    // We are going to iterate though it, and break it up in blocks.
    // We dont't always break on exactly maxColumnHeight, instead we try
    // to find a natural spot NEAR the maxColumnHeight, such as at the end
    // of a paragraph or at the end of a sentence.

    block currentBlock;
    blocks allBlocks;
    for(auto lineIt = in.begin(); lineIt != in.end(); ++lineIt)
    {
        std::wstring line = *lineIt;
        std::wstring nextLine;
        auto itNext = lineIt + 1;
        if (itNext != in.end())
        {
            nextLine = *itNext;
        }
        // The block needs to be exactly columnWidth wide to ensure alignment
        // is correct when we add the hspacing + next column. If the line is 
        // shorter, we have to pad with spaces.
        while (line.length() < columnWidth)
        {
            line += L" ";
        }
        currentBlock.emplace_back(line);

        // How far away from the bottom (=max height)?
        const int distanceToMaxHeight = maxColumnHeight - currentBlock.size();

        // If we have more than maxColumnHeight lines, it's time to start a new
        // block no matter what.
        if (distanceToMaxHeight <= 0)
        {
            allBlocks.emplace_back(currentBlock);
            currentBlock.clear();
            continue;
        }

        // Otherwise, are we past 2/3 down?
        if (distanceToMaxHeight < (static_cast<int>(maxColumnHeight) / 3))
        {
            // If the current or next line starts a new paragraph, let's just
            // start a new block here.
            if (line.empty() || nextLine.empty())
            {
                allBlocks.emplace_back(currentBlock);
                currentBlock.clear();
                continue;
            }
        }

        // Still here? ok... are we 1/4th of the way down?
        if (distanceToMaxHeight < (static_cast<int>(maxColumnHeight) / 4))
        {
            // We're almost to the end, we should probably start looking for a
            // good spot to break now if we haven't found a new paragraph yet.
            // Any line ending with punctuation would be good.

            const wchar_t lastChar = line[line.size() - 1];
            if (std::any_of(blockEnders.begin(), blockEnders.end(),
                [&lastChar](wchar_t c){
                    return c == lastChar;
                }))
            {
                allBlocks.emplace_back(currentBlock);
                currentBlock.clear();
                continue;
            }
        }

        // If we're still here, we'll just keep iterating until we hit the
        // maximum column height.
    }
    // We might have some left-over.
    if (!currentBlock.empty()) allBlocks.emplace_back(currentBlock);

    // Cool! we have blocks now - now we have to arrange them into ROWS.
    // Each row will have N columns where
    //     N = width - columnHorizontalSpacing / columnWidth
    blocks::iterator curBlockIt = allBlocks.begin();
    while (curBlockIt != allBlocks.end())
    {
        // Make a vector with all the blocks that will make up the current row.
        blocks blocksThisRow;
        for (size_t i = 0 ; i != numCols; ++i)
        {
            blocksThisRow.emplace_back(*curBlockIt);
            ++curBlockIt;
            if (curBlockIt == allBlocks.end()) break;
        }
        // Figure out how tall this row is going to be (how many lines), by
        // looking for the biggest block int the vector we just made.
        size_t curRowHeight{0};
        for (auto& aBlock: blocksThisRow)
        {
            if (aBlock.size() > curRowHeight) curRowHeight = aBlock.size();
        }
        // We also add the vertical spacing here to append blank rows in the
        // loop below.
        curRowHeight += columnVerticalSpacing;
        // For every line (0 to curRowHeight), grab a piece out of each block.
        std::wstring outLine;
        for (size_t nLine = 0; nLine != curRowHeight; ++nLine)
        {
            for (auto& iBlock: blocksThisRow)
            {
                if (iBlock.size() < nLine + 1)
                {
                    // If this block has no content for this line,
                    // fill with spaces instead.
                    outLine.append(columnWidth + columnHorizontalSpacing, ' ');
                }
                else
                {
                    outLine += iBlock[nLine];
                    outLine.append(columnHorizontalSpacing, ' ');
                }
            }
            // We'll have leftover spacing on the right, trim it off.
            outLine = StringRightTrim(outLine);
            result.emplace_back(outLine);
            outLine.clear();
        }
        // If desired, add a page separator line using the desired characters.
        if (pageSepLineChar != 0)
        {
            result.emplace_back(std::wstring(width - 1, pageSepLineChar));
        }
    }
    return result;
}
