/*
   Copyright© 2022 John Sennesael

   This file is part of justify.
   Justify is free software: you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the 
   Free Software Foundation, either version 3 of the License, 
   or (at your option) any later version.

   Justify is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with Justify. If not, see <https://www.gnu.org/licenses/>.
*/

#include "justify.h"

#include "columns.h"
#include "justifyCenter.h"
#include "justifyFill.h"
#include "justifyLeft.h"
#include "justifyRight.h"
#include "stringHelper.h"

#include <sys/ioctl.h>
#include <unistd.h>

#include <cstdlib>
#include <fstream>
#include <iostream>
#include <memory>
#include <string>
#include <vector>

void usage(const std::string& program)
{

    std::cout << std::endl;
    std::cout << "justify - simple text alignment tool." << std::endl;
    std::cout << "Copyright© 2021 John Sennesael" << std::endl << std::endl;
    std::cout << "Usage: " << program << " [OPTION]..." << std::endl;

    std::cout << std::endl << "Input / output options:" 
              << std::endl << std::endl;
    std::cout << "  -i,--input [FILE]     Input file to process, defaults to "
              << "stdin." << std::endl;
    std::cout << "  -o,--output [FILE]    Output file, defaults to stdout." 
              << std::endl;

    std::cout << std::endl << "Column options:" << std::endl << std::endl;
    std::cout << "  -c,--cols [COLS]      Number of columns for multi-column"
              << std::endl
              << "                        output." << std::endl;
    std::cout << "  -s,--hspacing [COLS]  Column horizontal spacing."
              << std::endl
              << "                        (default: 2)" << std::endl;
    std::cout << "  -v,--vspacing [ROWS]  Column vertical spacing."
              << std::endl
              << "                        (default: 1)" << std::endl;
    std::cout << "  -m,--colheight [ROWS] Column max. height."
              << std::endl
              << "                        (default: 0, =infinite)"
              << std::endl;
    std::cout << "  -p,--pagesep [CHAR]   Page separator line character."
              << std::endl;
    std::cout << std::endl << "Alignment options: "
              << std::endl << std::endl;
    std::cout << "  -w,--width [NUM]      Text width to format text into."
              << std::endl
              << "                        (defaults to terminal width)"
              << std::endl;
    std::cout << "  -a,--align [center|fill|left|right]" << std::endl
              << "                        Justify method (default=fill)."
              << std::endl;

    std::cout << std::endl << "Misc options: " << std::endl << std::endl;
    std::cout << "  -h,--help             Show this help text."
              << std::endl << std::endl;
}

bool checkParam(
    const std::string& paramType,
    const std::string& arg,
    const StrVector& args,
    StrVector::const_iterator& it)
{
    ++it;
    if (it == args.end())
    {
        std::cerr << "Missing argument to " << arg << "." << std::endl;
        std::cerr << "Expected: " << paramType << "." << std::endl;
        usage(args[0]);
        return false;
    }
    return true;
}

Settings parseArgs(const StrVector& args)
{
    Settings result{};
    result.error = true;
    for (auto it = args.begin() + 1; it != args.end(); ++it)
    {
        std::string arg = *it;
        if ((arg == "-i") || (arg == "--input"))
        {
            if (!checkParam("filename", arg, args, it)) return result;
            result.inFile = *it;
            continue;
        }
        else if ((arg == "-o") || (arg == "--output"))
        {
            if (!checkParam("filename", arg, args, it)) return result;
            result.outFile = *it;
            continue;
        }
        else if ((arg == "-c") || (arg == "--cols"))
        {
            if (!checkParam("number", arg, args, it)) return result;
            try
            {
                result.cols = std::stol(*it);
            }
            catch (const std::exception& e)
            {
                std::cerr << "Could not parse " << arg << " as a number: ";
                std::cerr << e.what() << std::endl;
                usage(args[0]);
                return result;
            }
            continue;
        }
        else if ((arg == "-s") || (arg == "--hspacing"))
        {
            if (!checkParam("number", arg, args, it)) return result;
            try
            {
                result.hspacing = std::stol(*it);
            }
            catch (const std::exception& e)
            {
                std::cerr << "Could not parse " << arg << " as a number: ";
                std::cerr << e.what() << std::endl;
                usage(args[0]);
                return result;
            }
            continue;
        }
        else if ((arg == "-v") || (arg == "--vspacing"))
        {
            if (!checkParam("number", arg, args, it)) return result;
            try
            {
                result.vspacing = std::stol(*it);
            }
            catch (const std::exception& e)
            {
                std::cerr << "Could not parse " << arg << " as a number: ";
                std::cerr << e.what() << std::endl;
                usage(args[0]);
                return result;
            }
            continue;
        }
        else if ((arg == "-m") || (arg == "--colheight"))
        {
            if (!checkParam("number", arg, args, it)) return result;
            try
            {
                result.maxColHeight = std::stol(*it);
            }
            catch (const std::exception& e)
            {
                std::cerr << "Could not parse " << arg << " as a number: ";
                std::cerr << e.what() << std::endl;
                usage(args[0]);
                return result;
            }
            continue;
        }
        else if ((arg == "-a") || (arg == "--align"))
        {
            if (!checkParam("justification type", arg, args, it)) 
                return result;
            const std::string jtype = StringToLower(*it);
            if (jtype == "center")
            {
                result.justify = Justification::CENTER;
            }
            else if (jtype == "fill")
            {
                result.justify = Justification::FILL;
            }
            else if (jtype == "left")
            {
                result.justify = Justification::LEFT;
            }
            else if (jtype == "right")
            {
                result.justify = Justification::RIGHT;
            }
            else
            {
                std::cerr << "Invalid justification type: " << jtype;
                usage(args[0]);
                return result;
            }
            continue;
        }
        else if ((arg == "-p") || (arg == "--pagesep"))
        {
            if (!checkParam("character", arg, args, it)) return result;
            std::wstring charStr = WideStringFromString(*it);
            if (!charStr.empty())
            {
                result.pageSepChar = charStr[0];
            }
        }
        else if ((arg == "-h") || (arg == "--help"))
        {
            usage(args[0]);
            result.error = false;
            result.shouldExit = true;
            return result;
        }
        else if ((arg == "-w") || (arg == "--width"))
        {
            if (!checkParam("number", arg, args, it)) return result;
            try
            {
                result.width = std::stol(*it);
            }
            catch (const std::exception& e)
            {
                std::cerr << "Could not parse " << arg << " as a number: ";
                std::cerr << e.what() << std::endl;
                usage(args[0]);
                return result;
            }
            continue;
        }
        else
        {
            std::cerr << "Unknown option: " << arg << std::endl;
            result.error = true;
            usage(args[0]);
            return result;
        }
    }
    result.error = false;
    return result;
}

bool justify(const Settings& settings)
{
    std::wifstream inputFile;
    std::wofstream outputFile;
    std::shared_ptr<std::wistream> inputStream;
    std::shared_ptr<std::wostream> outputStream;
    if (settings.inFile.empty())
    {
        inputStream = std::shared_ptr<std::wistream>(
            &std::wcin,
            [](std::wistream*) noexcept {}
        );
    }
    else
    {
        try
        {
            inputFile.open(settings.inFile, std::ifstream::in);
            inputStream = std::shared_ptr<std::wistream>(
                &inputFile,
                [](std::wistream*) noexcept {}
            );
        }
        catch (const std::exception& e)
        {
            std::cerr << "Could not open input file \"" 
                      << settings.inFile << "\": " << e.what() << std::endl;
            return false;
        }
    }
    if (settings.outFile.empty())
    {
        outputStream = std::shared_ptr<std::wostream>(
            &std::wcout,
            [](std::wostream*) noexcept {}
        );
    }
    else
    {
        try
        {
            outputFile.open(settings.outFile, std::ifstream::out);
            outputStream = std::shared_ptr<std::wostream>(
                &outputFile,
                [](std::wostream*) noexcept {}
            );
        }
        catch (const std::exception& e)
        {
            std::cerr << "Could not open output file \""
                      << settings.outFile << "\": " << e.what() << std::endl;
            return false;
        }
    }
    size_t justifyWidth = settings.width;
    if (settings.cols > 1)
    {
        if (settings.hspacing >= settings.width)
        {
            std::cerr << "Error: Horizontal column spacing too big for given "
                       << "width." << std::endl;
            return false;
        }
        // Figure out the width of a column if we have columns.
        justifyWidth = settings.width / settings.cols;
        justifyWidth -=  (settings.cols - 1) * settings.hspacing;
    }
    std::vector<std::wstring> justifiedText;
    switch (settings.justify)
    {
        case Justification::FILL:
            justifiedText = justifyFill(inputStream, justifyWidth);
            break;
        case Justification::CENTER:
            justifiedText = justifyCenter(inputStream, justifyWidth);
            break;
        case Justification::RIGHT:
            justifiedText = justifyRight(inputStream, justifyWidth);
            break;
        case Justification::LEFT:
            justifiedText = justifyLeft(inputStream, justifyWidth);
            break;
        default:
            std::cerr << "Not implemented." << std::endl;
            return false;
    };
    std::vector<std::wstring> output;
    if (settings.cols > 1)
    {
        // unless overridden by the user, the column height will be the number
        // of justified lines divided by the number of desired columns.
        size_t maxColHeight = settings.maxColHeight;
        if (maxColHeight == 0)
        {
            maxColHeight = justifiedText.size() / settings.cols;
        }
        // Split justified text up into columns.
        output = columns(
            justifiedText,
            settings.cols,
            justifyWidth,
            settings.hspacing,
            settings.vspacing,
            maxColHeight,
            settings.width,
            settings.pageSepChar
        );
    }
    else
    {
        output = justifiedText;
    }
    // Print the result to the output stream.
    for (auto& line: output)
    {
        *outputStream << line << std::endl;
    }
    return true;
}

int main(int argc, char* argv[])
{

    // Set locale to make unicode input work properly.
    const char* langEnv = std::getenv("LANG");
    const std::string sysLocale{langEnv ? langEnv : "C"};
    try
    {
        std::locale::global(std::locale(sysLocale));
    } catch (const std::exception& e)
    {
        std::cerr << "Could not set locale to: " << sysLocale << " : " 
                  << e.what() << std::endl;
        return 5;
    }

    std::vector<std::string> arguments(argv, argv + argc);
    Settings settings  = parseArgs(arguments);
    if (settings.error) return 1;
    if (settings.shouldExit) return 0;
    if (settings.width == 0)
    {
        // Get terminal width (to use as a default)
        /// @TODO make sure this works on most unixes
        winsize termSize;
        ioctl(STDOUT_FILENO, TIOCGWINSZ, &termSize);
        settings.width = termSize.ws_col;
    }
    int retval{0};
    try {
        retval = justify(settings) == true ? 0 : 1;
    }
    catch (const std::exception& e)
    {
        std::cerr << "Error: " << e.what() << std::endl;
        retval = 2;
    }
    return retval;
}

