/*
   Copyright© 2022 John Sennesael

   This file is part of justify.
   Justify is free software: you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the 
   Free Software Foundation, either version 3 of the License, 
   or (at your option) any later version.

   Justify is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with Justify. If not, see <https://www.gnu.org/licenses/>.
*/

#include "justifyCommon.h"

#include "stringHelper.h"

#include <fstream>
#include <memory>
#include <stdexcept>
#include <string>
#include <vector>

std::pair<std::wstring, std::wstring> splitAtWidth(
    const std::wstring& input,
    size_t width)
{
    if (input.size() <= width) return std::make_pair(input, L"");
    const size_t cutPos = input.find_last_of(
        whitespace<wchar_t>(), 
        width
    );
    if ((cutPos == input.npos) || (cutPos > width))
    {
        throw std::runtime_error(
            "Encountered a line that could be not be split "
            "at given width. Try specifying a larger width."
        );
    }
    std::wstring truncated = input.substr(0, cutPos);
    std::wstring overflow = StringTrim(input.substr(cutPos)) + L" ";
    return std::make_pair(truncated, overflow);
}

std::vector<std::wstring> truncateLinesToFitWidth(
    std::shared_ptr<std::wistream> in,
    size_t width)
{
    std::vector<std::wstring> result;
    // Slurp input and arrange into a vector holding paragraphs.
    std::vector<std::vector<std::wstring>> paragraphs;
    std::vector<std::wstring> currentParagraph;
    while (true)
    {
        std::wstring line;
        if (!std::getline(*in, line))
        {
            break;
        }
        line = StringRightTrim(line);
        if (line.empty())
        {
            paragraphs.emplace_back(currentParagraph);
            currentParagraph.clear();
        }
        else
        {
            currentParagraph.emplace_back(line);
        }
    }
    // Truncate the lines in each paragraph to the correct width.
    std::wstring overflow;
    for (const auto& paragraph: paragraphs)
    {
        for (std::wstring line: paragraph)
        {
            line = overflow + line;
            std::wstring truncatedLine;
            std::tie(truncatedLine, overflow) = splitAtWidth(line, width);
            result.emplace_back(truncatedLine);
        }
        while (true)
        {
            if (overflow.empty() || overflow == L" ") break;
            std::wstring truncatedLine;
            std::tie(truncatedLine, overflow) = splitAtWidth(overflow, width);
            result.emplace_back(truncatedLine);
        }
        result.emplace_back(L"");
    }
    return result;
}

