/*
   Copyright© 2022 John Sennesael

   This file is part of justify.
   Justify is free software: you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the 
   Free Software Foundation, either version 3 of the License, 
   or (at your option) any later version.

   Justify is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with Justify. If not, see <https://www.gnu.org/licenses/>.
*/

#include "justifyFill.h"

#include "justifyCommon.h"
#include "stringHelper.h"

#include <array>
#include <fstream>
#include <iostream>
#include <memory>
#include <string>
#include <vector>

std::vector<WordScore> calculateWordScores(
    const std::vector<std::wstring>& vWords)
{
    static std::array<wchar_t, 10> punctuation = {
        '!', '?', '.', ';', ':', '"', '\'', '}', ']', '>'
    };
    // Calculate average word length.
    float averageWordLength{0};
    std::wstring longestWord;
    for (std::wstring word: vWords)
    {
        averageWordLength += word.size();
        if (word.size() > longestWord.size())
        {
            longestWord = word;
        }
    }
    averageWordLength /= vWords.size();
    // Each word will get a score indicating on whether it's a good place to
    // insert a space.
    //
    // * Words ending in punctuation get a higher score for append.
    // * Words longer than the average word length get a higher score (because
    //   the distance to the next space would be larger)
    // * Word that is _THE_ longest word gets an extra point.
    // * Each space already in the word token detracts from the score.
    std::vector<WordScore> wordScores;
    for (auto it = vWords.begin(); it != vWords.end(); ++it)
    {
        const std::wstring wordToken = *it;
        const std::wstring word = StringTrim(wordToken);
        if (word.empty()) continue;
        const wchar_t lastChar = word[word.size() - 1];
        WordScore score{};
        score.iterator = it;
        score.score = 1;
        if (std::any_of(punctuation.begin(), punctuation.end(),
            [&lastChar](wchar_t c){
                return c == lastChar;
            }))
        {
            ++score.score;
        }
        if (word.size() > averageWordLength)
        {
            ++score.score;
        }
        if (word == longestWord)
        {
            ++score.score;
        }
        const size_t spacesCount = std::count(
            wordToken.begin(),
            wordToken.end(),
            ' '
        );
        if (spacesCount > 0)
        {
            score.score -= spacesCount;
        }
        wordScores.emplace_back(score);
    }
    return wordScores;
}

WordScore findHighestScore(const std::vector<WordScore>& vScores)
{
    if (vScores.empty())
    {
        throw std::runtime_error("Can't find the highest score without scores.");
    }
    WordScore highest = vScores.front();
    for (auto& score: vScores)
    {
        if (score.score > highest.score)
        {
            highest = score;
        }
    }
    return highest;
}

std::unique_ptr<WordScore> findScoreFor(
    const std::vector<WordScore>& scores,
    const std::vector<std::wstring>::const_iterator& searchFor)
{
    for (auto it = scores.begin(); it != scores.end(); ++it)
    {
        if ((*it).iterator == searchFor)
        {
            return std::make_unique<WordScore>((*it));
        }
    }
    return {};
}

bool invert{false};
std::vector<std::wstring> injectSpace(
    const std::vector<std::wstring>& wordsVector)
{
    std::vector<std::wstring> vWords = wordsVector;
    std::vector<std::wstring> result;
    // In order to avoid favoring the first highest score in the case of ties
    // every time, we flip the array upside down every other iteration, which
    // ensures a more even distribution of spaces.
    if (invert)
    {
        std::reverse(vWords.begin(), vWords.end());
    }

    if (vWords.empty()) return vWords;
    const auto wordScores = calculateWordScores(vWords);
    if (wordScores.empty()) 
    {
        // This happens for instance if there's only 1 word.
        throw std::runtime_error("No scores could be produced.");
    }
    auto winner = findHighestScore(wordScores).iterator;

    for (auto it = vWords.begin(); it != vWords.end(); ++it)
    {
        const auto found = findScoreFor(wordScores, it);
        if (found)
        {
            if (it == winner)
            {
                result.emplace_back((*it) + L" ");
            }
            else
            {
                result.emplace_back(*it);
            }
        }
    }

    // We we inversed the vector at the start, we have to invert it again before
    // returning.
    if (invert)
    {
        std::reverse(result.begin(), result.end());
    }
    invert = !invert;
    return result;
}

std::vector<std::wstring> justifyFill(
    std::shared_ptr<std::wistream> in,
    size_t width)
{
    std::vector<std::wstring> result;

    std::vector<std::wstring> parsedLines = truncateLinesToFitWidth(
        in,
        width
    );

    // Insert spaces until line width matches desired width.
    for (auto it = parsedLines.begin(); it != parsedLines.end(); ++it)
    {
        std::wstring line = StringRemoveDuplicateWhitespace(*it);
        std::wstring nextLine = L"NONE";
        auto itNext = it + 1;
        if (itNext != parsedLines.end())
        {
            nextLine = *itNext;
        }
        const size_t spacesNeeded = width - line.length();
        if (spacesNeeded == 0)
        {
            result.emplace_back(line);
            continue;
        }
        auto vWords = words(line);
        const size_t nWords = vWords.size();
        if (nWords < 1)
        {
            result.emplace_back(line);
            continue;
        }
        const size_t spacesPerWord = (spacesNeeded + nWords - 1) / nWords;

        // avoid injecting an unreasonable amount of whitespace.
        if (spacesPerWord > nWords)
        {
            result.emplace_back(line);
            continue;
        }

        // we need at least 2 words for this to work.
        if (nWords < 2)
        {
            result.emplace_back(line);
            continue;
        }

        // If this is the last line in a paragraph, and if the line isn't too
        // long, don't justify.
        if (nextLine.empty() && (line.size() <= width))
        {
            result.emplace_back(line);
            continue;
        }

        while (line.size() < width)
        {
            const auto startSize = line.size();
            auto vWordsWithoutLastWord = vWords;
            auto lastWordIt = --vWordsWithoutLastWord.end();
            const auto lastWord = *lastWordIt;
            vWordsWithoutLastWord.erase(lastWordIt);
            try
            {
                vWordsWithoutLastWord = injectSpace(vWordsWithoutLastWord);
            }
            catch(const std::exception&)
            {
                break;
            }
            vWords = vWordsWithoutLastWord;
            vWords.emplace_back(lastWord);
            line = StringJoin<wchar_t>(vWords, L" ");
            line = StringRightTrim(line);
            const auto endSize = line.size();
            if (startSize + 1 != endSize)
            {
                break;
            }
        }
        result.emplace_back(line);
    }
    return result;
}
