/*
   Copyright© 2022 John Sennesael

   This file is part of justify.
   Justify is free software: you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the 
   Free Software Foundation, either version 3 of the License, 
   or (at your option) any later version.

   Justify is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with Justify. If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef STRINGHELPER_H_INCLUDE
#define STRINGHELPER_H_INCLUDE

#include <algorithm>
#include <codecvt>
#include <locale>
#include <string>
#include <vector>

template <typename T>
struct IsWideCharType
{
    static const bool VALUE = false;
};

template <>
struct IsWideCharType<wchar_t>
{
    static const bool VALUE = true;
};

template <typename T, bool isWideChar>
struct WhiteSpaceImpl;

template <typename T>
struct WhiteSpaceImpl<T, false>
{
    static const char* get()
    {
        return " \t\n\r\f\v";
    }
};

template <typename T>
struct WhiteSpaceImpl<T, true>
{
    static const wchar_t* get()
    {
        return L" \t\n\r\f\v";
    }
};

template <typename T>
const T* whitespace()
{
   return WhiteSpaceImpl<T, IsWideCharType<T>::VALUE>::get();
}

template <typename T>
std::vector<std::basic_string<T>> words(const std::basic_string<T>& str)
{
    size_t pos=0;
    std::basic_string<T> s = str;
    std::vector<std::basic_string<T>> result;
    if (str.empty()) return result;
    int tokens = 1;
    while ((pos = s.find_first_of(whitespace<T>())) 
        != std::basic_string<T>::npos)
    {
        result.push_back(s.substr(0, pos));
        s.erase(0, pos + 1);
        tokens++;
    }
    result.push_back(s);
    return result;
}

template<typename T>
std::basic_string<T> StringLeftTrim(
    const std::basic_string<T>& str,
    const T* t = whitespace<T>())
{
    std::basic_string<T> s{str};
    s.erase(0, s.find_first_not_of(t));
    return s;
}

template<typename T>
std::basic_string<T> StringRightTrim(
    const std::basic_string<T>& str,
    const T* t = whitespace<T>())
{
    std::basic_string<T> s{str};
    s.erase(s.find_last_not_of(t) + 1);
    return s;
}

template<typename T>
std::basic_string<T> StringTrim(
    const std::basic_string<T>& str,
    const T* t = whitespace<T>())
{
    std::basic_string<T> s{str};
    return StringLeftTrim<T>(StringRightTrim<T>(s, t), t);
}

template<typename T>
std::basic_string<T> StringRemoveDuplicateWhitespace(
    const std::basic_string<T>& input)
{
    std::basic_string<T> result;
    std::unique_copy(input.begin(), input.end(), 
        std::back_insert_iterator<std::basic_string<T>>(result),
        [](char a,char b){
            return std::isspace(a) && std::isspace(b);
        }
    );
    return result;
}

template<typename T>
std::basic_string<T> StringJoin(
    const std::vector<std::basic_string<T>>& list,
    const std::basic_string<T>& delim)
{
    std::basic_string<T> result;
    if (list.empty()) return result;
    if (list.size() == 1) return list[0];
    for (const auto& token: list)
    {
        result += token;
        result += delim;
    }
    return result.substr(0, result.size() - delim.size());
}

template<typename T>
std::basic_string<T> StringToLower(const std::basic_string<T>& str)
{
    std::basic_string<T> copy = str;
    std::transform(copy.begin(), copy.end(), copy.begin(), ::tolower);
    return copy;
}

std::wstring WideStringFromString(const std::string& input);

#endif
